"""Contains classes representing melds that check their validity."""
from __future__ import annotations
from enum import Flag, auto
from itertools import combinations
from typing import Optional, Sequence, Union, Iterable, List, Iterator, TypeVar, Type, Tuple
from .tiles import Honors, Tile, Simples, Bonuses, Misc, Wind

__all__ = [
    'WuFlag',
    'Meld',
    'Pong',
    'Kong',
    'Chow',
    'Eyes',
    'Wu',
]

class WuFlag(Flag):
    CHICKEN_HAND = 0

    # types
    COMMON_HAND = 1 << 0
    ALL_IN_TRIPLETS = 1 << 1
    MIXED_ONE_SUIT = 1 << 2
    ALL_ONE_SUIT = 1 << 3
    ALL_HONOR_TILES = 1 << 4
    SMALL_DRAGONS = 1 << 5
    GREAT_DRAGONS = 1 << 6
    SMALL_WINDS = 1 << 7
    GREAT_WINDS = 1 << 8
    NINE_GATES = 1 << 9
    ALL_KONGS = 1 << 10
    SELF_TRIPLETS = 1 << 11
    ORPHANS = 1 << 12
    THIRTEEN_ORPHANS = 1 << 13

    # presence of certain tiles
    SEAT_WIND = 1 << 14
    PREVAILING_WIND = 1 << 15
    RED_DRAGON = 1 << 16
    GREEN_DRAGON = 1 << 17
    WHITE_DRAGON = 1 << 18
    MIXED_ORPHANS = 1 << 19

    # winning condition
    SELF_DRAW = 1 << 20
    ALL_FROM_WALL = 1 << 21
    ROBBING_KONG = 1 << 22
    LAST_CATCH = 1 << 23
    BY_KONG = 1 << 24
    DOUBLE_KONG = 1 << 25
    HEAVENLY = 1 << 26
    EARTHLY = 1 << 27

    # penalties
    TWELVE_PIECE = 1 << 28
    GAVE_DRAGON = 1 << 29
    GAVE_KONG = 1 << 30

    # bonuses
    NO_BONUSES = 1 << 31
    ALIGNED_FLOWERS = 1 << 32
    ALIGNED_SEASONS = 1 << 33
    TABLE_OF_FLOWERS = 1 << 34
    TABLE_OF_SEASONS = 1 << 35
    HAND_OF_BONUSES = 1 << 36

# special case: 1 & 9 of each suit + every value of honors suits
THIRTEEN_ORPHANS = set(map(Tile.from_str, (
    'tong/1|tong/9|zhu/1|zhu/9|wan/1|wan/9|'  # simples
    'feng/1|feng/2|feng/3|feng/4|long/1|long/2|long/3'  # honors
).split('|')))

FLAG_FAAN = {
    WuFlag.CHICKEN_HAND: 0,
    WuFlag.COMMON_HAND: 1, WuFlag.ALL_IN_TRIPLETS: 3,
    WuFlag.ALL_HONOR_TILES: 10, WuFlag.ALL_ONE_SUIT: 7,
    WuFlag.MIXED_ONE_SUIT: 3, WuFlag.GREAT_DRAGONS: 8,
    WuFlag.SMALL_DRAGONS: 5, WuFlag.GREAT_WINDS: 13,
    WuFlag.SMALL_WINDS: 10, WuFlag.THIRTEEN_ORPHANS: 13,
    WuFlag.ALL_KONGS: 13, WuFlag.SELF_TRIPLETS: 8,
    WuFlag.ORPHANS: 10, WuFlag.NINE_GATES: 10,
    WuFlag.RED_DRAGON: 1, WuFlag.GREAT_WINDS: 1,
    WuFlag.WHITE_DRAGON: 1, WuFlag.SEAT_WIND: 1,
    WuFlag.PREVAILING_WIND: 1, WuFlag.MIXED_ORPHANS: 1,
    WuFlag.SELF_DRAW: 1, WuFlag.ALL_FROM_WALL: 1,
    WuFlag.ROBBING_KONG: 1, WuFlag.LAST_CATCH: 1,
    WuFlag.BY_KONG: 1, WuFlag.DOUBLE_KONG: 8,
    WuFlag.HEAVENLY: 13, WuFlag.EARTHLY: 13,
    WuFlag.TWELVE_PIECE: 0, WuFlag.GAVE_DRAGON: 0, WuFlag.GAVE_KONG: 0,
    WuFlag.NO_BONUSES: 1, WuFlag.ALIGNED_FLOWERS: 1,
    WuFlag.ALIGNED_SEASONS: 1, WuFlag.TABLE_OF_FLOWERS: 2,
    WuFlag.TABLE_OF_SEASONS: 2, WuFlag.HAND_OF_BONUSES: 8,
}

def _tname(obj) -> str:
    return type(obj).__name__

T = TypeVar('T', bound='Meld')

class Meld:
    """Represents a meld of tiles."""

    @property
    def size(self) -> Union[int, range]:
        """Valid size of this meld."""
        raise NotImplementedError

    tiles: Tuple[Tile, ...]

    def __init__(self, tiles: Iterable[Tile]):
        """Check validity upon initialization."""
        self.tiles = tuple(sorted(tiles))
        self.check_meld()

    @classmethod
    def from_str(cls: Type[T], s: str) -> T:
        """Meld.from_str('suit1/num1|...') -> Meld"""
        return cls(map(Tile.from_str, s.split('|')))

    def __str__(self) -> str:
        """str(meld) -> 'suit1/num1|suit2/num2|...'"""
        tiles = list(self.tiles)
        if isinstance(self, Wu):
            for meld in self.fixed_melds:
                tiles.extend(meld.tiles)
        tiles.sort()
        return '|'.join(map(str, tiles))

    __repr__ = __str__

    def __hash__(self) -> int:
        return hash((type(self),) + self.tiles)

    def __eq__(self, other: Meld) -> bool:
        """Two Melds are equal if they are same type and same tiles."""
        if not isinstance(other, Meld):
            return NotImplemented
        return hash(self) == hash(other)

    def __lt__(self, other: Meld) -> bool:
        order = [Wu, Kong, Pong, Chow, Eyes]
        if type(self) is not type(other):
            return order.index(type(self)) > order.index(type(other))
        return self.tiles < other.tiles

    def check_meld(self) -> None:
        """Check validity of the meld. Raises ValueError upon failure."""
        self.check_size()
        for tile in self.tiles:
            if isinstance(tile.suit, Bonuses):
                raise ValueError('No Bonus tiles in a meld')

    def check_size(self) -> None:
        """Check that the meld is the right size."""
        if isinstance(self.size, range) and isinstance(self, Wu):
            size = len(self.tiles + tuple(
                tile for meld in self.fixed_melds
                for tile in meld.tiles))
            if size not in self.size:
                raise ValueError(f'{_tname(self)} must be '
                                 f'[{self.size.start},{self.size.stop}) tiles')
        else:
            if len(self.tiles) != self.size:
                raise ValueError(f'{_tname(self)} must be {self.size} tiles')

    def check_suit(self) -> None:
        """Check that all tiles are the same suit."""
        suit = self.tiles[0].suit
        for tile in self.tiles[1:]:
            if tile.suit != suit:
                raise ValueError(
                    f'{_tname(self)} must be all one suit: '
                    f'{suit} != {tile.suit}'
                )

class _SameNum(Meld):
    """Base class that provides check_num method."""

    def check_meld(self) -> None:
        """Check validity of the meld."""
        super().check_meld()
        self.check_num()

    def check_num(self) -> None:
        """Check that all tiles are the same number/value."""
        self.check_suit()
        num = self.tiles[0].number
        for tile in self.tiles[1:]:
            if tile.number != num:
                raise ValueError(
                    f'{_tname(self)}s must be the same tile: '
                    f'{num} != {tile.number}'
                )

class Pong(_SameNum):
    """Represents a Pong (three identical tiles)"""

    size: int = 3

class Kong(_SameNum):
    """Represents a Kong (four identical tiles, counted as three)"""

    size: int = 4

    def __str__(self) -> str:
        """Kongs are represented by one faceup, two stacked down, one faceup"""
        #pylint: disable=no-member
        return '|'.join(map(str, [
            self.tiles[0], Misc.HIDDEN.value, self.tiles[-1]]))
        #pylint: enable=no-member

class Chow(Meld):
    """Represents a Chow (three tiles of the same suit with consecutive numbers)"""

    size: int = 3

    def check_meld(self) -> None:
        """Check validity as a Chow."""
        super().check_meld()
        self.check_suit()
        if not isinstance(self.tiles[0].suit, Simples):
            raise ValueError(
                f'Chows can only be Simples, not {_tname(self.tiles[0].suit)}')
        num = self.tiles[0].number
        for i, tile in enumerate(self.tiles):
            if tile.number != num + i:
                raise ValueError('Chows must have consecutive numbers: '
                                 f'{tile.number} != {num + i}')

class Eyes(_SameNum):
    """Represents a pair of Eyes (two identical tiles, only valid in winning hand)"""

    size: int = 2

class Wu(Meld):
    """Represents a winning hand. This can only consist of sub-melds.

    WARNING: for some hands, instantiating this class is NOT atomic.
    So far I have discovered hands that take a full second or two.
    Consider instantiating this class a blocking procedure, and
    treat it as such, e.g. in asynchronous contexts.
    """

    size = range(14, 19) # [14, 18]
    melds: List[List[Meld]]
    fixed_melds: List[Meld]
    arrived: Optional[Tile]
    discarder: Optional[Wind]
    base_flags: WuFlag

    @property
    def all_tiles(self) -> List[Tile]:
        tiles = list(self.tiles)
        for meld in self.fixed_melds:
            tiles.extend(meld.tiles)
        tiles.sort()
        return tiles

    def __init__(self, tiles: Iterable[Tile], melds: Iterable[Meld] = None,
                 arrived: Tile = None, discarder: Optional[int] = None,
                 flags: WuFlag = WuFlag.CHICKEN_HAND):
        """Check validity upon initialization."""
        self.tiles = tuple(sorted(tiles))
        self.fixed_melds = list(melds or [])
        self.arrived = arrived
        self.discarder = Wind(discarder) if discarder is not None else None
        self.base_flags = flags
        self.check_meld()

    @classmethod
    def from_str(cls: Type[Wu], s: str, *a, **kw) -> Wu:
        """Wu.from_str('suit1/num1|...') -> Wu"""
        return cls(map(Tile.from_str, s.split('|')), *a, *kw)

    def check_meld(self) -> None:
        """Check that this hand is a winning hand.
        Returns all possible winning combinations.
        """
        super().check_meld()
        # sort each combo of melds, to make out-of-order duplicates
        # equal, then pass them through a set to weed them out
        self.melds = list(map(list, set(
            tuple(sorted(combo)) for combo in self.valid_combos())))
        if not self.melds:
            raise ValueError('No valid combos')

    def valid_combos(self) -> Iterator[List[Meld]]:
        """Yield all possible winning sets of melds."""
        tc = len(self.tiles)
        checked = set()
        for e1, e2 in combinations(range(tc), 2):
            if e1 == e2:
                continue
            try:
                test_eyes = Eyes((self.tiles[e1], self.tiles[e2]))
            except ValueError:
                continue
            if test_eyes in checked:
                continue
            checked.add(test_eyes)
            # tiles other than the eyes we're checking
            ts = [x for i, x in enumerate(self.tiles) if i != e1 and i != e2]
            possible = self.all_melds_pos(ts)
            combos = combinations(possible, 4 - len(self.fixed_melds))
            for combo in combos:
                covered = {i for meld, idxs in combo for i in idxs}
                if len(covered) != len(ts):
                    continue
                covered = [tile for meld, idxs in combo for tile in meld.tiles]
                if len(covered) != len(ts):
                    continue
                melds = [meld for meld, idxs in combo] + list(self.fixed_melds)
                melds.append(test_eyes)
                yield melds
        if set(self.tiles) == THIRTEEN_ORPHANS:
            melds: List[Meld] = []
            melds.append(_UncheckedWu(self.tiles))
            yield melds

    @staticmethod
    def get_triplet(trip: List[Tile]) -> Union[Meld, None]:
        """Try to get a meld."""
        try:
            return Pong(trip)
        except ValueError:
            pass
        try:
            return Chow(trip)
        except ValueError:
            pass
        return None

    def all_melds_pos(self, ts) -> List[Tuple[Meld, Tuple[int, int, int]]]:
        """Get every possible meld and the indexes of their component cards.
        Modified from
        https://github.com/offe/py-mcr/blob/master/mahjonggrouping.py#L108-L121
        """
        trips = {}
        tc = len(ts)
        for t1, t2, t3 in combinations(range(tc), 3):
            t1, t2, t3 = trip_idxs = tuple(sorted((t1, t2, t3)))
            # already checked
            if trip_idxs in trips:
                continue
            # not unique indexes
            if len(set(trip_idxs)) != 3:
                continue
            trip = [ts[i] for i in trip_idxs]
            meld = self.get_triplet(trip)
            if isinstance(meld, Pong) and t3 < tc - 1:
                old_idxs = trip_idxs
                old_meld = meld
                for t4 in range(t3 + 1, tc):
                    trip_idxs = tuple(sorted((t1, t2, t3, t4)))
                    if trip_idxs in trips or len(set(trip_idxs)) != 4:
                        continue
                    trip = [ts[i] for i in trip_idxs]
                    try:
                        meld = Kong(trip)
                    except ValueError:
                        continue
                    else:
                        break
                else:
                    trip_idxs = old_idxs
                    meld = old_meld
            if meld:
                trips[trip_idxs] = meld
        return sorted((i, j) for j, i in trips.items())

    def flags(self, choice: Sequence[Meld],
             winds: Tuple[int, int] = None) -> WuFlag:
        """WuFlags that apply to this choice of winning hand."""
        all_tiles = self.all_tiles
        types = self.base_flags
        if all(isinstance(meld, (Chow, Eyes)) for meld in choice):
            types |= WuFlag.COMMON_HAND
        if all(isinstance(meld, (Pong, Kong, Eyes)) for meld in choice):
            types |= WuFlag.ALL_IN_TRIPLETS
        suit = None
        hon = False
        all_one_suit = False
        for tile in all_tiles:
            if isinstance(tile.suit, Honors):
                hon = True
                continue
            if suit is not None and tile.suit != suit:
                break
            suit = tile.suit
        else:
            if hon and suit is None:
                # no regulars, only honors
                types |= WuFlag.ALL_HONOR_TILES
            elif not hon:
                # no honors, only regulars
                types |= WuFlag.ALL_ONE_SUIT
            else:
                # regulars (but of one suit) and honors
                types |= WuFlag.MIXED_ONE_SUIT
        long: List[Optional[type]] = [None, None, None]
        for meld in choice:
            if isinstance(meld, Chow):
                continue # can't be dragons
            if meld.tiles[0].suit != Honors.LONG:
                continue # not dragons
            long[meld.tiles[0].number] = type(meld)
        if all(long):
            if not any(typ is Eyes for typ in long):
                types |= WuFlag.GREAT_DRAGONS
            else:
                types |= WuFlag.SMALL_DRAGONS
        feng: List[Optional[type]] = [None, None, None, None]
        for meld in choice:
            if isinstance(meld, Chow):
                continue # can't be winds
            if meld.tiles[0].suit != Honors.FENG:
                continue # not winds
            feng[meld.tiles[0].number] = type(meld)
        if all(feng):
            if not any(typ is Eyes for typ in feng):
                types |= WuFlag.GREAT_WINDS
            else:
                types |= WuFlag.SMALL_WINDS
        if len(choice) == 1: # only one possible way this can happen
            types |= WuFlag.THIRTEEN_ORPHANS
        if all(isinstance(meld, (Kong, Eyes)) for meld in choice):
            types |= WuFlag.ALL_KONGS
        if not self.fixed_melds and WuFlag.ALL_IN_TRIPLETS in types:
            if WuFlag.SELF_DRAW in types or ( # self draw or
                self.arrived is not None
                # stolen discard not in regular melds,
                and all(self.arrived not in meld
                        for meld in choice
                        if not isinstance(meld, Eyes))
                # only eyes
                and all(self.arrived in meld.tiles
                        for meld in choice
                        if isinstance(meld, Eyes))
            ):
                types |= WuFlag.SELF_TRIPLETS
        for meld in choice:
            if isinstance(meld, Chow):
                break
            if not isinstance(meld.tiles[0].suit, Simples):
                break
            if meld.tiles[0].number not in {0, 8}:
                break
        else:
            types |= WuFlag.ORPHANS
        counts = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        goals = [3, 1, 1, 1, 1, 1, 1, 1, 3]
        suit = all_tiles[0].suit
        if all_one_suit and isinstance(suit, Simples):
            for tile in all_tiles:
                counts[tile.number] += 1
            diff = 0
            for i in range(9):
                if counts[i] == goals[i]:
                    continue
                if counts[i] == goals[i] + 1:
                    diff += 1
                    continue
                break
            else:
                if diff == 1:
                    types |= WuFlag.NINE_GATES
        favorable = list({
            'long/1': WuFlag.RED_DRAGON,
            'long/2': WuFlag.GREEN_DRAGON,
            'long/3': WuFlag.WHITE_DRAGON
        }.items())
        if winds is not None:
            favorable.append((f'feng/{winds[0]+1}', WuFlag.SEAT_WIND))
            favorable.append((f'feng/{winds[1]+1}', WuFlag.PREVAILING_WIND))
        for meld in choice:
            for s, flag in favorable:
                if meld.tiles[0] == Tile.from_str(s):
                    types |= flag
        for tile in all_tiles:
            if isinstance(tile.suit, Honors):
                continue
            if tile.number in {0, 8}:
                continue
            break
        else:
            types |= WuFlag.MIXED_ORPHANS
        if not self.fixed_melds:
            types |= WuFlag.ALL_FROM_WALL
        # unset flags
        if WuFlag.SMALL_WINDS in types:
            types &= ~(WuFlag.SEAT_WIND | WuFlag.PREVAILING_WIND)
        if WuFlag.SELF_TRIPLETS in types:
            types &= ~(WuFlag.ALL_FROM_WALL | WuFlag.ALL_IN_TRIPLETS)
        if WuFlag.ALL_HONOR_TILES in types:
            types &= ~(WuFlag.ALL_IN_TRIPLETS)
        if WuFlag.ORPHANS in types:
            types &= ~(WuFlag.ALL_IN_TRIPLETS)
        if WuFlag.NINE_GATES in types:
            types &= ~(WuFlag.ALL_ONE_SUIT | WuFlag.ALL_FROM_WALL)
        if WuFlag.THIRTEEN_ORPHANS in types:
            types &= ~(WuFlag.MIXED_ORPHANS | WuFlag.ORPHANS)
        return types

    def faan(self, choice: Sequence[Meld],
              winds: Tuple[int, int] = None) -> Tuple[int, WuFlag]:
        """Number of faan points the Wu is worth.

        Parameters
        -----------
        ``choice``: Sequence[Meld]
            Which combination of melds from the possible
            set to calculate the points for.
        ``winds``: Optional[Tuple[int, int]]
            If specified as (seat, prevailing), this provides the necessary
            context to include the Seat Wind and Prevailing Wind Wu flags.

        Returns
        --------
        Tuple[int, WuFlag]
            (points, flags): flags represents all of the features of a Wu
            that are true for this particular combination.
        """
        types = self.flags(choice, winds)
        return (faan(types), types)

    def max_faan(self, winds: Tuple[int, int] = None) \
        -> Tuple[List[Meld], int, WuFlag]:
        return max(
            ((choice, *self.faan(choice, winds)) for choice in self.melds),
            key=lambda trip: trip[1])

def faan(flags: WuFlag) -> int:
    points = 0
    for flag in WuFlag:
        if flag in flags:
            points += FLAG_FAAN[flag]
    return points

class _UncheckedWu(Wu):
    size = range(13, 19)

    def __init__(
        self, tiles: Iterable[Tile], melds: Iterable[Meld] = None, *_, **__
    ):
        """Don't check validity, that's the whole point"""
        self.tiles = tuple(sorted(tiles)) + tuple(
            tile for meld in (melds or []) for tile in meld.tiles)
        self.fixed_melds = []

if 0:
    # TODO: figure out why one particular one takes multiple seconds:
    # 1.25s (3.94s when debug): tong/1|tong/1|tong/1|tong/2|tong/2|tong/2|tong/3|tong/3|tong/3|tong/4|tong/4|tong/4|tong/5|tong/5
    # Note that fixing even one meld reduces it to atomic:
    # 0.02s: tong/1|tong/1|tong/1 + tong/2|tong/2|tong/2|tong/3|tong/3|tong/3|tong/4|tong/4|tong/4|tong/5|tong/5
    # Other timing tests follow
    # 0.00s: tong/1|tong/9|zhu/1|zhu/9|wan/1|wan/9|feng/1|feng/2|feng/3|feng/4|long/1|long/2|long/3|wan/2
    # 0.08s: wan/1|wan/1|wan/1|wan/2|wan/3|wan/4|wan/4|wan/5|wan/6|wan/7|wan/8|wan/9|wan/9|wan/9
    # 0.18s: tong/1|tong/2|tong/3|tong/2|tong/3|tong/4|tong/5|tong/6|tong/7|tong/8|tong/8|tong/8|tong/4|tong/4
    # 0.32s: feng/1|feng/1|feng/1|feng/1|feng/2|feng/2|feng/2|feng/2|feng/3|feng/3|feng/3|feng/3|feng/4|feng/4|feng/4|feng/4|long/1|long/1
    # 0.17s: tong/1|tong/1|tong/1|wan/9|wan/9|wan/9|zhu/5|zhu/5|zhu/5|tong/2|tong/2|tong/2|long/1|long/1
    # 0.00s: zhu/3|zhu/4|zhu/5,zhu/5|zhu/6|zhu/7,zhu/7|zhu/8|zhu/9 + feng/3|feng/3|zhu/4|zhu/5|zhu/6
    # 0.55s: wan/2|wan/2|wan/2|tong/1|tong/1|tong/1|tong/1|tong/2|tong/3|tong/2|tong/2|tong/2|tong/3|tong/3
    pass



''' # comment out this opening triple quote to time Wu computation
import mahjong, time
start = time.time()
w = mahjong.melds.Wu.from_str("""
tong/4|tong/4|tong/4|tong/5|tong/6|tong/6|tong/6|tong/7
""".strip(),
melds=[mahjong.Pong.from_str('zhu/3|zhu/3|zhu/3'), mahjong.Chow.from_str('wan/1|wan/2|wan/3')],
arrived=mahjong.Tile.from_str('tong/4')
#map(mahjong.Chow.from_str, 'zhu/3|zhu/4|zhu/5,zhu/5|zhu/6|zhu/7,zhu/7|zhu/8|zhu/9'.split(','))
)
end = time.time()
print('\n'.join(f'{w.faan(meld)}: ' + ', '.join(str(i) for i in meld)
                for meld in set(tuple(meld) for meld in w.melds)))
print(w.max_faan())
print(f'in {end-start:.2f}s')
#'''
''' # comment out this opening triple quote to test dealing
from mahjong import Game, Round, Hand
game = Game()
round = Round(game)
hand = Hand(round)
hand.shuffle()
hand.deal()
for p in game.players:
    print('|'.join(map(str, p.hand)))
    print(', '.join(map(str, p.bonus)))
#'''
#''' # uncomment this opening triple quote when commenting out one elsewhere
import sys
from mahjong.game import Game, Hand, UserIO, Question, HandEnding, HandResult

if '--game' in sys.argv:
    game = Game(extra_hands=False)
else:
    game = Hand(None)
print('Note: all indexes are **1-based**.')
print('This is a rudimentary text-based mahjong implementation.')
print("It's purely as a proof-of-concept/manual testing method.")
print("It requires trust that each player won't look at the other's privacy.")
print("Please don't actually use this. Make something even a tiny bit better.")
print("With that said, let's start.")

question = game.play()
while question is not None:
    if isinstance(question, UserIO):
        if question.question == Question.READY_Q:
            input('Hit Enter when ready for the next round.')
            question = question.answer()
            continue
        hand_minus_tile = [tile for tile in question.hand if tile is not question.arrived]
        print('Question for Player #%s' % question.player.seat.value)
        print('Draw/Last discard: %s;' % question.arrived, 'Concealed:',
              ', '.join(map(str, hand_minus_tile)))
        print('Shown:', '\t'.join(map(str, question.shown)),
              '- Bonuses:', '|'.join(map(str, question.player.bonus)))
        if question.question == Question.DISCARD_WHAT:
            idx = int(input('Enter index of card to discard. 0 for draw. '))
            if idx == 0:
                if question.arrived in question.hand:
                    tile = question.arrived
                else:
                    tile = hand_minus_tile[0]
            elif idx > 0:
                tile = hand_minus_tile[idx-1]
            elif idx < 0:
                tile = hand_minus_tile[idx]
            else:
                tile = hand_minus_tile[0]
            print('Discarding', tile)
            question = question.answer(tile)
        elif question.question == Question.MELD_FROM_DISCARD_Q:
            assert question.melds is not None
            print('Available melds to meld:', '\t'.join(map(str, question.melds)))
            try:
                idx = int(input('Enter index of meld to meld, or blank to not meld: '))
            except ValueError:
                question = question.answer(None)
            else:
                question = question.answer(question.melds[idx-1])
        elif question.question == Question.SHOW_EKFCP_Q:
            assert question.melds is not None
            print('Available Kongs to expose from concealed Pongs:',
                  '\t'.join(map(str, question.melds)))
            try:
                idx = int(input('Enter index of Kong to expose, or blank to not expose: '))
            except ValueError:
                question = question.answer(None)
            else:
                question = question.answer(question.melds[idx-1])
        elif question.question == Question.SHOW_EKFEP_Q:
            assert question.melds is not None
            print('Available Kongs to expose from exposed Pongs:',
                  '\t'.join(map(str, question.melds)))
            try:
                idx = int(input('Enter index of Kong to expose, or blank to not expose: '))
            except ValueError:
                question = question.answer(None)
            else:
                question = question.answer(question.melds[idx-1])
        elif question.question == Question.ROB_KONG_Q:
            assert question.melds is not None
            print('You can rob the last Kong to win with:', question.melds[0])
            inp = '?'
            while inp not in {'y', 'n'}:
                inp = input('Do you want to? (y/n) ').casefold()
            question = question.answer(inp == 'y')
        elif question.question == Question.WHICH_WU:
            assert question.melds is not None
            idx = int(input('Which Wu combo (enter index) do you want to win with? '))
            question = question.answer(question.melds[idx-1])
    elif isinstance(question, HandEnding):
        if question.result == HandResult.GOULASH:
            print('Goulash! Nobody wins. Starting next game...')
        else:
            assert question.choice is not None
            print('Player #%s won with %s (%s faan; %s points; %s)! Starting next game...' % (
                question.winner.seat.value, ','.join(map(str, question.choice)),
                question.faan()[0], *question.points(1)
            ))
        question = question.answer()
print('Game Over!')
#'''



from typing import List, Tuple
from .tiles import Bonuses, Tile, BonusTile, Wind
from .melds import Meld, WuFlag, faan

class Player:
    """Represents one Mahjong player."""

    seat: Wind
    hand: List[Tile]
    shown: List[Meld]
    bonus: List[BonusTile]

    @property
    def full_hand(self):
        """All tiles in this player's possession, except Bonuses."""
        return sorted(self.hand + [tile for meld in self.shown
                                   for tile in meld.tiles])

    def __init__(self, seat: int):
        self.seat = Wind(seat)
        self.hand = []
        self.shown = []
        self.bonus = []

    def __repr__(self) -> str:
        return f'<Player #{self.seat}: {",".join(map(repr, self.hand))}, '\
            f'shown {self.shown}>'

    __str__ = __repr__

    def draw(self, wall: List[Tile]) -> Tile:
        """Draw a tile from the wall, and keep doing so if it's a Bonus."""
        tile = wall.pop()
        while isinstance(tile, BonusTile):
            self.bonus.append(tile) # Step 21
            self.bonus.sort()
            tile = wall.pop()
        self.hand.append(tile) # Step 20
        return tile

    def show_meld(self, discard: Tile, meld: Meld):
        """Save a meld and remove its tiles from the hand."""
        self.hand.append(discard)
        for tile in meld.tiles:
            self.hand.remove(tile)
        self.shown.append(meld) # Step 19

    def bonus_faan(self) -> Tuple[int, WuFlag]:
        """Get faan won from bonus tiles or lack thereof."""
        flags = WuFlag.CHICKEN_HAND
        if not self.bonus:
            flags |= WuFlag.NO_BONUSES
        found = {
            Bonuses.GUI: [False]*4,
            Bonuses.HUA: [False]*4
        }
        for tile in self.bonus:
            if tile.number == self.seat:
                if tile.suit == Bonuses.HUA:
                    flags |= WuFlag.ALIGNED_FLOWERS
                elif tile.suit == Bonuses.GUI:
                    flags |= WuFlag.ALIGNED_SEASONS
            found[tile.suit][tile.number] = True
        for k, v in found.items():
            if all(v):
                if k == Bonuses.HUA:
                    flags |= WuFlag.TABLE_OF_FLOWERS
                elif k == Bonuses.GUI:
                    flags |= WuFlag.TABLE_OF_SEASONS
        if all(map(all, found.values())):
            flags = WuFlag.HAND_OF_BONUSES
        if WuFlag.TABLE_OF_FLOWERS in flags:
            flags &= ~(WuFlag.ALIGNED_FLOWERS)
        if WuFlag.TABLE_OF_SEASONS in flags:
            flags &= ~(WuFlag.ALIGNED_SEASONS)
        return (faan(flags), flags)


"""Enums and the Tile class."""
from __future__ import annotations
from enum import Enum, IntEnum
from typing import Union

__all__ = [
    'Suit',
    'Simples',
    'Honors',
    'Bonuses',
    'Misc',
    'Wind',
    'Dragon',
    'Flower',
    'Season',
    'ORDER',
    'Number',

    'Tile',
    'BonusTile',
]

class Suit(Enum):
    """Base enum class for suits."""
    pass

class Simples(Suit):
    """Enum for simples suits (numbered from 1-9)"""
    ZHU = 'zhu'
    TONG = 'tong'
    WAN = 'wan'

class Honors(Suit):
    """Enum for honors suits (tiles that cannot Chow)"""
    FENG = 'feng'  # value is instance of Wind
    LONG = 'long'  # value is instance of Dragon

class Bonuses(Suit):
    """Enum for bonus suits (tiles that only count when you win)"""
    HUA = 'hua'
    GUI = 'gui'

class Misc(Suit):
    """Enum for special case suits"""
    UNKNOWN = '?'
    HIDDEN = 'k'
    MISSING = ' '

class Wind(IntEnum):
    """Represents a wind direction"""
    EAST = 0
    SOUTH = 1
    WEST = 2
    NORTH = 3

class Dragon(IntEnum):
    """Represents a dragon"""
    RED = 0 # hong zhong
    GREEN = 1 # fa cai
    WHITE = 2 # bai ban

class Flower(IntEnum):
    """Represents a Flowers tile"""
    MEI = Wind.EAST
    LAN = Wind.SOUTH
    JU = Wind.WEST
    ZHU = Wind.NORTH

class Season(IntEnum):
    """Represents a Seasons tile"""
    SPRING = Wind.EAST
    SUMMER = Wind.SOUTH
    AUTUMN = Wind.WEST
    WINTER = Wind.NORTH

ORDER = {
    'wan': 0, 'tong': 1, 'zhu': 2,
    'feng': 3, 'long': 4
}

Number = Union[int, Wind, Dragon, Flower, Season]

class Tile:
    """Data class for tiles."""
    suit: Suit
    number: Number

    def __init__(self, suit: Suit, number: Number):
        """Initialize Tile."""
        self.suit = suit
        self.number = number

    @classmethod
    def from_str(cls, s: str) -> Tile:
        """Tile.from_str('suit/number') -> Tile"""
        suit_s, number_s = s.split('/')
        suit: Suit = Misc.UNKNOWN
        for num in (Simples, Honors, Bonuses, Misc):
            try:
                suit = num(suit_s)
            except ValueError:
                continue
            else:
                break
        else:
            raise ValueError('Invalid suit')
        number: int = int(number_s) - 1
        if not (0 <= number < 9):
            raise ValueError(f'Number {number_s} not in range [1, 9]')
        if suit == Honors.FENG:
            number = Wind(number)
        if suit == Honors.LONG:
            number = Dragon(number)
        return cls(suit, number)

    def __str__(self) -> str:
        """str(tile) -> 'suit/number'"""
        return f'{self.suit.value}/{self.number}'

    __repr__ = __str__

    def __hash__(self):
        return hash((self.suit, self.number))

    def __eq__(self, other: Tile) -> bool:
        """Tiles are equal when their suits and numbers are equal."""
        if not (isinstance(other, type(self)) or isinstance(self, type(other))):
            return NotImplemented
        return hash(self) == hash(other)

    def __lt__(self, other: Tile) -> bool:
        if not (isinstance(other, type(self)) or isinstance(self, type(other))):
            return NotImplemented
        if self.suit == other.suit:
            return self.number < other.number
        if self.suit.value not in ORDER or other.suit.value not in ORDER:
            return False
        return ORDER[self.suit.value] < ORDER[other.suit.value]

class BonusTile(Tile):
    suit: Bonuses
    number: Union[int, Flower, Season]

    def __init__(self, suit: Bonuses, number: Union[int, Flower, Season]):
        """Initialize Tile."""
        self.suit = suit
        self.number = number



"""All game-process-related classes."""
from __future__ import annotations
from enum import Enum
from typing import Generator, List, Mapping, NamedTuple, Optional, \
    Tuple, Union, Dict, overload
from itertools import combinations
import random
from .tiles import BonusTile, Bonuses, Honors, Simples, Tile, Wind
from .melds import Chow, Kong, Meld, Pong, Wu, WuFlag
from .players import Player

# This is a really big file, but there's no way around it:
# All of the classes depend on each other, so there's
# a lot of circular imports if they are in different files.

__all__ = [
    'TurnEnding',
    'HandEnding',
    'Question',
    'UserIO',
    'HandResult',

    'Game',
    'Round',
    'Hand',
    'Turn',
]

STOCK_TABLES = {
    'enwp': {
        1: 1, 2: 1, 3: 1,
        4: 2, 5: 2, 6: 2,
        7: 4, 8: 4, 9: 4,
        10: 8
    },
    'zhwp': {
        0: 1, 1: 2, 2: 4, 3: 8,
        4: 16, 5: 24, 6: 32, 7: 48,
        8: 64, 9: 96, 10: 128, 11: 192,
        12: 256, 13: 384
    },
    # in the app I play,
    # 3 faan = 1200c = 1x1200
    # 4 faan = 2400c = 2x1200
    # 5 faan = 3600c = 3x1200
    # 6 faan = 4800c = 4x1200
    # 7 faan = 7200c = 6x1200
    # 8 faan = 9600c = 8x1200
    'random_app': {
        1: 1, 2: 1, 3: 1,
        4: 2, 5: 3, 6: 4,
        7: 6, 8: 8, 9: 10,
        10: 13
    }
}

# data classes

def _answer(gen: Generator, ans=None) -> YieldType:
    """Send the answer to the internal generator."""
    try:
        return gen.send(ans)
    except StopIteration:
        return None

class TurnEnding(NamedTuple):
    winner: Optional[Player] = None
    discard: Optional[Tile] = None
    seat: Optional[Wind] = None
    wu: Optional[Wu] = None
    jumped_with: Optional[Meld] = None
    offered: bool = False
    prev_seat: Optional[Wind] = None

class HandEnding(NamedTuple):
    result: HandResult
    hand: Hand
    winner: Optional[Player] = None
    wu: Optional[Wu] = None
    choice: Optional[List[Meld]] = None

    def answer(self, ans=None):
        return _answer(self.hand.gen, ans)

    def faan(self) -> Tuple[int, Optional[WuFlag]]:
        """Calculate the faan for this hand."""
        if self.winner is None or self.wu is None or self.choice is None:
            return (0, None)
        points = self.wu.faan(self.choice, (self.winner.seat, self.hand.wind))
        bonus = self.winner.bonus_faan()
        return (points[0] + bonus[0], points[1] | bonus[1])

    def points(self, min_faan: int = 3, table: Mapping[int, int] = None) \
            -> Tuple[List[int], Optional[WuFlag]]:
        """Calculate the change in points for each player.

        Parameters
        -----------
        ``min_faan``: int
            Minimum faan before the hand is recognized as a valid winning hand
        ``table``: Mapping[int, int]
            Mapping of faan to base points. The largest faan value is assumed
            to be the limit, and values larger than it will be mapped to it.

        Returns
        --------
        Tuple[List[int], Optional[WuFlag]]
            A list of four point deltas, representing each player, and
            the flags that apply to this hand. The former is four zeroes,
            and the latter is None, on goulash or not enough faan.
        """
        points = [0, 0, 0, 0]
        if self.winner is None:
            return (points, None)
        faan, flags = self.faan()
        if faan < min_faan:
            return (points, None)
        winner = self.winner
        wu = self.wu
        if table is None:
            table = STOCK_TABLES['random_app']
        limit = max(table.keys())
        base = table[min(faan, limit)]
        # special penalties that make the perpetrator pay
        # for everyone else on top of themselves
        if ((WuFlag.TWELVE_PIECE in wu.flags and WuFlag.SELF_DRAW in wu.flags)
            or (WuFlag.GAVE_KONG in wu.flags and WuFlag.SELF_DRAW in wu.flags)
            or WuFlag.GAVE_DRAGON in wu.flags) and wu.discarder is not None:
            points[winner.seat] += base * 3
            points[wu.discarder] -= base * 3
        # loser pays double in normal losing conditions
        elif wu.discarder is not None:
            points[winner.seat] += base * 2
            points[wu.discarder] -= base * 2
        # self draw means everyone pays
        elif WuFlag.SELF_DRAW in wu.flags:
            for i in range(4):
                if i == winner.seat:
                    points[i] += base * 3
                else:
                    points[i] -= base
        else:
            raise ValueError('Somehow, nobody gets points. '
                             'Possibly report to maintainer with '
                             'the following information: '
                             + str(self.wu.__dict__)
                             + ' ;; ' + str(self.winner.__dict__))
        return (points, flags)

class Question(Enum):
    MELD_FROM_DISCARD_Q = 16
    SHOW_EKFCP_Q = 23
    SHOW_EKFEP_Q = 24
    DISCARD_WHAT = 27
    ROB_KONG_Q = 33
    WHICH_WU = 40
    READY_Q = 0
    SELF_DRAW_Q = 13

class UserIO(NamedTuple):
    question: Question
    gen: Generator
    melds: Union[List[Meld], List[List[Meld]], None] = None
    player: Optional[Player] = None
    arrived: Optional[Tile] = None
    last_meld: Optional[Meld] = None

    @property
    def hand(self) -> List[Tile]:
        """Shortcut to .player.hand"""
        return self.player.hand

    @property
    def shown(self) -> List[Meld]:
        """Shortcut to .player.shown"""
        return self.player.shown

    def answer(self, ans=None):
        return _answer(self.gen, ans)

    def __repr__(self) -> str:
        return f'UserIO(question={self.question}, melds={self.melds}, '\
            f'arrived={self.arrived}, player={self.player})'

class HandResult(Enum):
    """The result of a hand."""
    NORMAL = 0
    GOULASH = 1
    DEALER_WON = 2

YieldType = Union[UserIO, HandEnding, None]

# game process classes

class Game:
    """Represents an entire game (consisting of 4 Rounds)."""

    round: Round

    def __init__(self, **house_rules):
        self.init_players()
        self.extra_hands = house_rules.pop('extra_hands', True)

    def play(self) -> Union[UserIO, HandEnding]:
        """Play the game!

        Starts and stores a generator.
        Returns the first request for user input.
        """
        if hasattr(self, 'gen'):
            raise RuntimeError('Game already started!')
        self.gen = self.execute()
        question = next(self.gen)
        return question

    def execute(self):
        """Generator-coroutine-based interface to play the game."""
        for i in range(4): # Step 01, 03
            self.round = Round(self)
            yield from self.round.execute(i) # Step 02
            yield UserIO(Question.READY_Q, self.gen)


    def init_players(self: Union[Game, Hand]) -> None:
        """Setup players."""
        self.players = [Player(i) for i in range(4)]

class Round:
    """Represents one four(+)-hand round of Mahjong."""

    hand: Hand
    game: Game
    wind: Wind

    def __init__(self, game: Game):
        self.game = game

    def execute(self, wind: int):
        """Play at least four hands."""
        self.wind = Wind(wind)
        i = int(self.wind)
        j = 0
        while j < 4: # Step 07
            self.hand = Hand(self)
            result: HandEnding = (yield from self.hand.execute(i)) # Step 05
            if result.result is HandResult.NORMAL: # Step 06
                i += 1 # Step 04
                j += 1
            elif not self.game.extra_hands: # house rules
                i += 1
                j += 1
            i %= 4
            yield result

class Hand:
    """Represents one full hand of Mahjong."""

    wind: Wind
    turn: Turn
    round: Optional[Round]
    wall: List[Tile]
    discarded: List[Tile]
    ending: Optional[TurnEnding] = None
    turncount: Optional[int] = None

    # key got last meld from value, and then self drew
    _12_piece: Dict[Player, Player]
    # key got last dragon meld from value, and then self drew
    _gave_dragon: Dict[Player, Player]
    # key got kong from value, and then self drew from the kong
    _gave_kong: Dict[Player, Player]

    @overload
    def __init__(self, round: Round):
        """Setup a hand.

        Pass the :class:`Round`_ explicitly if you have control over Round
        machinery *and* Game machinery (it will reach into the round's game
        attribute to get the game's players).

        Otherwise, this is the default invocation by the Round machinery.
        """

    @overload
    def __init__(self, players: List[Player]):
        """Setup a hand.

        Pass ``[p1, p2, p3, p4]`` if you want more control over how hands
        are played and have control over the players list.
        """

    @overload
    def __init__(self, standalone: None):
        """Setup a hand.

        Pass ``None`` if this hand is intended to be played standalone.
        The players will be initialized automagically.
        """

    def __init__(self, round_or_players: Union[Round, List[Player], None]):
        """Setup a hand."""
        if isinstance(round_or_players, Round):
            self.round = round_or_players
            self.players = self.round.game.players
        elif isinstance(round_or_players, list):
            self.round = None
            self.players = round_or_players
        else:
            self.round = None
            Game.init_players(self)
        self._12_piece = {}
        self._gave_dragon = {}
        self._gave_kong = {}

    def play(self, wind: Wind = Wind.EAST) -> Union[UserIO, HandEnding]:
        """Play the hand standalone.

        ``wind``: The prevailing wind for this hand.

        Starts and stores a generator.
        Returns the first request for user input.
        """
        if hasattr(self, 'gen'):
            raise RuntimeError('Game already started!')
        self.gen = self.execute(wind)
        try:
            question = next(self.gen)
        except StopIteration as exc:
            return exc.value
        if isinstance(question, UserIO):
            return question
        raise RuntimeError('This should never happen. Contact the developer.')

    def execute(self, wind: int):
        """Play a hand till it ends."""
        if not hasattr(self, 'gen') and self.round is not None:
            self.gen = self.round.game.gen
        self.wind = Wind(wind)
        self.shuffle() # Step 08
        self.deal() # Step 09
        # Sets up Step 12
        ending: TurnEnding = TurnEnding(seat=self.wind)
        self.turncount = 0
        while ending.winner is None and self.wall:
            self.turn = Turn(self)
            # Step 13
            ending = (yield from self.turn.execute(ending, self.turncount))
            self.turncount += 1
        self.ending = ending
        # Step 15
        if ending.winner is None: # implies and not self.wall
            # tie
            return HandEnding(HandResult.GOULASH, self)
        if len(self.wall) <= 0:
            ending.wu.base_flags |= WuFlag.LAST_CATCH
        if ending.winner in self._gave_dragon:
            ending.wu.base_flags |= WuFlag.GAVE_DRAGON
            ending.wu.discarder = self._gave_dragon[ending.winner].seat
        if WuFlag.SELF_DRAW in ending.wu.flags \
                and ending.winner in self._12_piece:
            ending.wu.base_flags |= WuFlag.TWELVE_PIECE
            ending.wu.discarder = self._12_piece[ending.winner].seat
        if len(ending.wu.melds) > 1:
            question = UserIO(Question.WHICH_WU, self.gen,
                              melds=ending.wu.melds)
            answer = (yield question)
        else:
            answer = ending.wu.melds[0]
        if ending.winner is self.players[self.wind]:
            return HandEnding(
                HandResult.DEALER_WON, self,
                ending.winner, ending.wu, answer)
        return HandEnding(
            HandResult.NORMAL, self, ending.winner, ending.wu, answer)

    def shuffle(self):
        """Generate and shuffle a new wall."""
        wall = []
        # simples
        for suit in Simples:
            for num in range(9):
                wall.extend([Tile(suit, num) for _ in range(4)])
        # honors
        for wind in range(4):
            wall.extend([Tile(Honors.FENG, wind) for _ in range(4)])
        for dragon in range(3):
            wall.extend([Tile(Honors.LONG, dragon) for _ in range(4)])
        # bonuses
        for i in range(4):
            wall.append(BonusTile(Bonuses.HUA, i))
            wall.append(BonusTile(Bonuses.GUI, i))
        random.shuffle(wall)
        self.wall = wall
        self.discarded = []

    def deal(self):
        players = self.players
        for player in players:
            player.__init__(player.seat)
        # Step 09
        for _ in range(13):
            for player in players:
                player.draw(self.wall)
        for player in players:
            player.hand.sort()

class Turn:
    """Handles the turn."""

    hand: Hand

    def __init__(self, hand: Hand):
        self.hand = hand
        self.players = self.hand.players
        self.gen = self.hand.gen

    def execute(self, last_ending: TurnEnding, turncount: int):
        if last_ending.seat is None:
            raise ValueError('Must have valid seat')
        player = self.players[last_ending.seat]
        meld: Optional[Meld] = None
        if last_ending.jumped_with is not None \
                and last_ending.discard is not None \
                and last_ending.prev_seat is not None:
            meld = last_ending.jumped_with
            self.hand.discarded.pop() # Step 18
            old_count = len(player.shown)
            old_drag = self.all_dragons(player)
            player.show_meld(last_ending.discard, meld)
            if old_count == 3 and len(player.shown) == 4:
                self.hand._12_piece[player] = self.players[
                    last_ending.prev_seat]
            if not old_drag and self.all_dragons(player):
                self.hand._gave_dragon[player] = self.players[
                    last_ending.prev_seat]
            del old_count, old_drag
        elif last_ending.discard is not None and not last_ending.offered:
            # Step 16
            meld: Optional[Meld] = (yield from self.melds_from_discard(
                player, last_ending))
        if isinstance(meld, Wu):
            if turncount == 1:
                meld.base_flags |= WuFlag.EARTHLY
            return TurnEnding(winner=player, wu=meld)
        if isinstance(meld, Kong):
            wu, winner = (yield from self.check_kong_robbers(meld, player))
            if wu is not None:
                return TurnEnding(winner=winner, wu=wu)
        if isinstance(meld, Kong) or meld is None:
            arrived = None
            kong: Optional[Kong] = None
            tries = 0
            wu, winner = None, None
            while 1:
                try:
                    drawn = player.draw(self.hand.wall) # Step 20
                except IndexError: # goulash time
                    return TurnEnding()
                try:
                    flags = WuFlag.SELF_DRAW
                    if tries == 0 and turncount == 0:
                        flags |= WuFlag.HEAVENLY
                    elif tries == 1:
                        flags |= WuFlag.BY_KONG
                    elif tries > 1:
                        flags |= WuFlag.DOUBLE_KONG
                    if (WuFlag.BY_KONG in flags or WuFlag.DOUBLE_KONG in flags) \
                            and player in self.hand._gave_kong:
                        flags |= WuFlag.GAVE_KONG
                        discarder = self.hand._gave_kong[player].seat
                    else:
                        discarder = None
                    wu = Wu(player.hand, player.shown, drawn, discarder, flags)
                except ValueError:
                    pass
                else:
                    question = UserIO(Question.SELF_DRAW_Q, self.gen,
                                      [wu], player, drawn)
                    ans: bool = (yield question)
                    if ans:
                        return TurnEnding(winner=player, wu=wu)
                    del ans
                arrived = drawn
                kong = (yield from self.check_ekfp(drawn, player))
                if kong is None:
                    self.hand._gave_kong.pop(player, None)
                    break
                player.shown.append(kong)
                wu, winner = (yield from self.check_kong_robbers(kong, player))
                if wu is not None:
                    return TurnEnding(winner=winner, wu=wu)
                tries += 1
            del tries, flags, kong, wu, winner
        else:
            arrived = last_ending.discard
        player.hand.sort() # Step 26
        question = UserIO(Question.DISCARD_WHAT, self.gen,
                          player=player, arrived=arrived, last_meld=meld)
        answer: Tile = (yield question) # Step 27
        # Step 28
        player.hand.remove(answer)
        self.hand.discarded.append(answer)
        # Step 29
        thief, meld = (yield from self.check_others_melds(answer, player))
        if thief is None and isinstance(meld, bool):
            # Step 14
            return TurnEnding(
                discard=answer, offered=meld,
                seat=Wind((player.seat + 1) % 4), prev_seat=player.seat)
        # Step 31
        if isinstance(meld, Wu):
            return TurnEnding(winner=thief, wu=meld)
        return TurnEnding(discard=answer, seat=thief.seat,
                          jumped_with=meld, prev_seat=player.seat)

    @staticmethod
    def all_dragons(player: Player) -> bool:
        dragons = [False, False, False]
        for m in player.shown:
            if m.tiles[0].suit == Honors.LONG:
                dragons[m.tiles[0].number] = True
        return all(dragons)

    def melds_from_discard(self, player: Player, last_ending: TurnEnding):
        """Check for possible melds to make from the discard tile."""
        if last_ending.discard is None:
            return None
        melds: List[Meld] = []
        for pair in combinations(player.hand, 2):
            meld = Wu.get_triplet([*pair, last_ending.discard])
            if meld is not None and meld not in melds:
                melds.append(meld)
        for triplet in combinations(player.hand, 3):
            if all(tile == last_ending.discard for tile in triplet):
                # Exposed Kong
                meld = Kong([*triplet, last_ending.discard])
                if meld not in melds:
                    melds.append(meld)
        try:
            wu = Wu([*player.hand, last_ending.discard], player.shown,
                    last_ending.discard, last_ending.prev_seat)
        except ValueError:
            pass
        else:
            melds.append(wu)
        if melds:
            # Step 17
            question = UserIO(Question.MELD_FROM_DISCARD_Q, self.gen,
                              melds, player, last_ending.discard)
            answer: Optional[Meld] = (yield question)
            if answer is not None:
                self.hand.discarded.pop() # Step 18
                player.show_meld(last_ending.discard, answer)
            return answer
        return None

    def check_ekfp(self, draw: Tile, player: Player):
        """Check whether an Exposed Kong From (any) Pong is possible."""
        kongs: List[Meld] = []
        for quadruplet in combinations(player.hand, 4):
            if all(tile == quadruplet[0] for tile in quadruplet):
                # Exposed Kong From Concealed Pong
                kong = Kong(quadruplet)
                kongs.append(kong)
        if kongs:
            question = UserIO(Question.SHOW_EKFCP_Q, self.gen,
                              melds=kongs, player=player, arrived=draw)
            answer: Optional[Kong] = (yield question)
            if answer is not None:
                for tile in answer.tiles:
                    player.hand.remove(tile)
                return answer
            kongs = []
        for meld in player.shown:
            if not isinstance(meld, Pong):
                continue
            try:
                match = player.hand.index(meld.tiles[0])
            except ValueError:
                continue
            if meld.tiles[0] in player.hand:
                kong = Kong(meld.tiles + (player.hand[match],))
                kongs.append(kong)
        if kongs:
            question = UserIO(Question.SHOW_EKFEP_Q, self.gen,
                              melds=kongs, player=player, arrived=draw)
            answer: Optional[Kong] = (yield question)
            if answer is not None:
                player.shown.remove(Pong(answer.tiles[:3]))
                player.hand.remove(answer.tiles[0])
            return answer
        return None

    def check_kong_robbers(self, kong: Kong, victim: Player):
        """Check for anyone who could potentially rob the Kong for a win."""
        tile = kong.tiles[0] # any tile will do
        for p in self.players:
            if p is victim:
                continue
            try:
                wu = Wu(p.hand + [tile], p.shown, tile, victim.seat, WuFlag.ROBBING_KONG)
            except ValueError:
                continue
            question = UserIO(Question.ROB_KONG_Q, self.gen,
                              melds=[wu], player=p)
            answer: bool = (yield question)
            if answer:
                return (wu, p)
        return (None, None)

    def check_others_melds(self, discard: Tile, victim: Player):
        """Check for anyone who could potentially meld from the discard."""
        overall: Mapping[Player, List[Meld]] = {}
        for player in self.players:
            if player is victim:
                continue
            melds: List[Meld] = overall.setdefault(player, [])
            for pair in combinations(player.hand, 2):
                try:
                    meld = Pong([*pair, discard])
                except ValueError:
                    pass
                else:
                    if meld not in melds:
                        melds.append(meld)
                if player is self.players[(victim.seat + 1) % 4]:
                    try:
                        meld = Chow([*pair, discard])
                    except ValueError:
                        continue
                    if meld not in melds:
                        melds.append(meld)
            for triplet in combinations(player.hand, 3):
                try:
                    meld = Kong([*triplet, discard])
                except ValueError:
                    continue
                if meld not in melds:
                    melds.append(meld)
            try:
                wu = Wu([*player.hand, discard], player.shown, discard, victim.seat)
            except ValueError:
                pass
            else:
                if wu not in melds:
                    melds.append(wu)
        for melds in overall.values():
            melds.sort()
        pairs: List[Tuple[Player, Meld]] = [
            (p, melds[0]) for p, melds in overall.items() if melds]
        pairs.sort(key=lambda i: i[1])
        for player, meld in pairs:
            # Step 30, 32
            question = UserIO(
                Question.MELD_FROM_DISCARD_Q, self.gen, melds=overall[player],
                player=player, arrived=discard)
            answer: Optional[Meld] = (yield question)
            if answer is not None:
                if isinstance(answer, Kong) \
                        and discard not in self.hand.discarded[:-1]:
                    self.hand._gave_kong[player] = victim
                return (player, answer)
        return (None, bool(overall[self.players[(victim.seat + 1) % 4]]))



"""
Mahjong
-------

Abstract away the logic of Mahjong games (for oh, it is complicated!)
Only worry about the representation.

Example Usage
=============

.. code-block:: python

    >>> import mahjong
    >>> game = mahjong.Game()
    >>> question = game.play()
    >>> next_question = question.answer(...)
    >>> ...

For full usage see the docs:
https://github.com/Kenny2github/mahjong/wiki
"""

__version__ = "2.0.0rc1"
__all__ = [
    'TurnEnding',
    'HandEnding',
    'Question',
    'UserIO',
    'HandResult',

    'Game',
    'Round',
    'Hand',
    'Turn',

    'Suit',
    'Simples',
    'Honors',
    'Bonuses',
    'Misc',
    'Wind',
    'Dragon',
    'Flower',
    'Season',
    'ORDER',
    'Number',

    'Tile',
    'BonusTile',

    'WuFlag',
    'Meld',
    'Pong',
    'Kong',
    'Chow',
    'Eyes',
    'Wu',
]

from .players import *
from .melds import *
from .tiles import *
from .game import *



from setuptools import setup
import re

with open('mahjong/__init__.py') as f:
    content = f.read()
    longdesc = re.match(r'^"""([\s\S]+?)"""', content).group(1).strip()
    with open('README.rst', 'w') as rdme:
        rdme.write(longdesc)
    version = re.search(r'__version__\s*=\s*"([^"]+)"', content).group(1)
del f, rdme

setup(
    name="python-mahjong",
    version=version,
    description="Abstract away the logic of mahjong games.",
    long_description=longdesc,
    url="https://github.com/Kenny2github/mahjong",
    author="Ken Hilton",
    license="LGPLv3+",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Games/Entertainment :: Board Games',
        'Topic :: Games/Entertainment :: Turn Based Strategy',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='mahjong game engine',
    packages=['mahjong'],
    python_requires='>=3.7',
)
